#!/bin/sh -f
#
#
#___INFO__MARK_BEGIN__
##########################################################################
#
#  The Contents of this file are made available subject to the terms of
#  the Sun Industry Standards Source License Version 1.2
#
#  Sun Microsystems Inc., March, 2001
#
#
#  Sun Industry Standards Source License Version 1.2
#  =================================================
#  The contents of this file are subject to the Sun Industry Standards
#  Source License Version 1.2 (the "License"); You may not use this file
#  except in compliance with the License. You may obtain a copy of the
#  License at http://gridengine.sunsource.net/Gridengine_SISSL_license.html
#
#  Software provided under this License is provided on an "AS IS" basis,
#  WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING,
#  WITHOUT LIMITATION, WARRANTIES THAT THE SOFTWARE IS FREE OF DEFECTS,
#  MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE, OR NON-INFRINGING.
#  See the License for the specific provisions governing your rights and
#  obligations concerning the Software.
#
#  The Initial Developer of the Original Code is: Sun Microsystems, Inc.
#
#  Copyright: 2001 by Sun Microsystems, Inc.
#
#  All Rights Reserved.
#
##########################################################################
#___INFO__MARK_END__

# startup of PVM conforming with Grid Engine
# parallel environment interface
#
# usage: startpvm.sh [options] <pe_hostfile> <master_host> <pvm_root>
#
#        options are:
#                     -ep <path_enhancement>
#                      If this option is given following argument added as 
#
#                         * ep=<path_enhancement>
#
#                      before the first host in the hostfile which is 
#                      generated by starpvm.sh. This causes pvmd to
#                      search binaries also in these pathes. Refer to 
#                      pvmd(1).
#
#                     -catch_hostname 
#                      force use of hostname wrapper in $TMPDIR when starting mpirun   
#
#                     -catch_rsh
#                      force use of rsh wrapper in $TMPDIR when starting mpirun 
#
#                      This will enable a tight integration of the PVM
#                      jobs. This is done by using a wrapper around the
#                      default rsh calls of PVM, and delegating all the
#                      created files to the SGE supplied $TMPDIR.


# useful to control parameters passed to us  
echo $*


catch_rsh=0
catch_hostname=0
path_enhancement=""
while [ $# -gt 0 ]; do
   case "$1" in
      -catch_rsh)
         catch_rsh=1
         ;;
      -catch_hostname)
         catch_hostname=1
         ;;
      -ep)
         shift
         path_enhancement=$1
         ;;
      *)
         break;
         ;;
   esac
   shift
done

me=`basename $0`

# test number of args
if [ $# -lt 3 ]; then
   echo "$me: got wrong number of arguments" >&2
   exit 1
fi

# get arguments
pe_hostfile=$1
host=$2
PVM_ROOT=$3
export PVM_ROOT

# get PVM_ARCH by starting $PVM_ROOT/lib/pvmgetarch 
if [ ! -x $PVM_ROOT/lib/pvmgetarch ]; then
   echo "$me: can't execute $PVM_ROOT/lib/pvmgetarch" >&2
   exit 1
fi
PVM_ARCH=`$PVM_ROOT/lib/pvmgetarch`

# ensure we are able to exec our starter
if [ ! -x $SGE_ROOT/pvm/bin/$ARC/start_pvm ]; then
   echo "$me: can't execute $SGE_ROOT/pvm/bin/$ARC/start_pvm" >&2
   exit 1
fi

# ensure we are able to start pvmd
if [ ! -x $PVM_ROOT/lib/pvmd ]; then
   echo "$me: can't execute $PVM_ROOT/lib/pvmd" >&2
   exit 1
fi

# ensure pe_hostfile is readable
if [ ! -r $pe_hostfile ]; then
   echo "$me: can't read $pe_hostfile" >&2
   exit 1
fi

# create pvm_hostfile
# remove column with number of slots per queue
# pvm does not support them in this form
pvm_hostfile="$TMPDIR/hostfile"

# enhance the search path if requested
if [ "x$path_enhancement" != "x" ]; then 
   echo "* ep=$path_enhancement" >> $pvm_hostfile
fi

cut -f1 -d" " $pe_hostfile >> $pvm_hostfile

# PVM 3.4.4 knows a Virtual Machine IDs for supporting
# overlapping PVM instances of the same user
# PVM_VMID=$JOB_ID; export PVM_VMID
#
# annotation: if you want to use PVM_VMID in
# addition to the tight integration, be sure to
# set this also in your job script (besides the
# necessary PVM_TMP).

#
# Make script wrapper for 'rsh' available in jobs tmp dir
#
if [ $catch_rsh = 1 ]; then
   rsh_wrapper=$SGE_ROOT/pvm/rsh
   if [ ! -x $rsh_wrapper ]; then
      echo "$me: can't execute $rsh_wrapper" >&2
      echo "     maybe it resides at a file system not available at this machine" >&2
      exit 1
   fi

   rshcmd=rsh
   case "$ARC" in
      hp|hp10|hp11|hp11-64) rshcmd=remsh ;;
      *) ;;
   esac
   # note: This could also be done using rcp, ftp or s.th.
   #       else. We use a symbolic link since it is the
   #       cheapest in case of a shared filesystem
   #
   ln -s $rsh_wrapper $TMPDIR/$rshcmd

   PVM_RSH=rsh; export PVM_RSH
fi

#
# Make script wrapper for 'hostname' available in jobs tmp dir
#
if [ $catch_hostname = 1 ]; then
   hostname_wrapper=$SGE_ROOT/pvm/hostname
   if [ ! -x $hostname_wrapper ]; then
      echo "$me: can't execute $hostname_wrapper" >&2
      echo "     maybe it resides at a file system not available at this machine" >&2
      exit 1
   fi

   # note: This could also be done using rcp, ftp or s.th.
   #       else. We use a symbolic link since it is the
   #       cheapest in case of a shared filesystem
   #
   ln -s $hostname_wrapper $TMPDIR/hostname
fi

# startup and wait for daemons
if [ $catch_rsh = 1 ]; then
   export PVM_TMP=$TMPDIR
   $SGE_ROOT/pvm/bin/$ARC/start_pvm -h $NHOSTS -n $host -c $PVM_ROOT/lib/pvmd $pvm_hostfile
else
   $SGE_ROOT/pvm/bin/$ARC/start_pvm -h $NHOSTS $PVM_ROOT/lib/pvmd $pvm_hostfile
fi

# on failure we call our cleanup skript
if [ $? -ne 0 ]; then
   echo "$me: startup failed - invoking cleanup script"

   if [ $catch_rsh = 1 ]; then
      $SGE_ROOT/pvm/stoppvm.sh -catch_rsh $pe_hostfile $host
   else
      $SGE_ROOT/pvm/stoppvm.sh $pe_hostfile $host
   fi
   exit 1
fi

# signal success to caller
exit 0
